# Copyright 2008, Michael Conrad Tadpol Tilstra <tadpol@tadpol.org>
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    


# source kitten.kit

package require tdom
package require tablelist

##############################################################################
# Model
variable ::dataAll

proc load_data {filename} {
	set fo [load_file $filename]
	set name [file tail $filename]

	dict set ::dataAll $name [timeAndDistance $fo]
	dict set ::dataAll $name Path $filename

	return $name
}

proc load_file {name} {
	set fp [open $name]
	fconfigure $fp -encoding utf-8
	set data [read $fp]
	# remove the namespace; it causes tdom to barf.
	regsub {xmlns="http://www.topografix.com/GPX/1/0"} $data "" data
	set doc [dom parse $data]
	set data [$doc documentElement root]
	close $fp

	foreach node [$data selectNodes {//trkpt}] {
		set time [string trim [[$node selectNodes time] text]]
		set lat [$node getAttribute lat]
		set lon [$node getAttribute lon]
		lappend ret [list $time $lat $lon]
	}
	return $ret
}

proc ::tcl::mathfunc::deg2rad {deg} {
	return [expr $deg * 3.14159265 / 180]
}
proc ::tcl::mathfunc::sinsq {val} {
	return [expr sin($val) * sin($val)]
}

proc timeAndDistance {data} {
	# convert lat&lon to distances
	set tDist 0
	lassign [lindex $data 0] pTime pLat pLon
	foreach point [lrange $data 1 end] {
		lassign $point time lat lon

		# Calculate distance from previous point via Haversine formula
		set dlat [expr deg2rad($lat - $pLat)]
		set dlon [expr deg2rad($lon - $pLon)]
		set a [expr sinsq($dlat/2) + cos($pLat) * cos($lat) * sinsq($dlon/2)]
		set dist [expr 6371 * (2 * atan2( sqrt($a), sqrt(1-$a) ))]

		# convert km to miles
		set dist [expr $dist * 0.621371192]

		# How much time elapsed?
		set tsec [clock scan $time -format %Y-%m-%dT%TZ -timezone :UTC]
		set psec [clock scan $pTime -format %Y-%m-%dT%TZ -timezone :UTC]
		set dur [expr abs($tsec - $psec)]

		# How fast did we go? (mph)
		set speed [expr $dist / ( $dur / 3600.0 )]

		lappend route [list $time $lat $lon $dist $dur $speed]
		set pTime $time
		set pLat $lat
		set pLon $lon

		set tDist [expr $tDist + $dist]
	}

	set tsec [clock scan [lindex $route end 0] -format %Y-%m-%dT%TZ -timezone :UTC]
	set psec [clock scan [lindex $route 0 0] -format %Y-%m-%dT%TZ -timezone :UTC]

	dict set ret Start [clock format $psec -format %T]
	dict set ret Stop [clock format $tsec -format %T]
	dict set ret Durration [expr ($tsec - $psec) / 60.0]
	dict set ret Distance $tDist
	dict set ret Speed [expr $tDist / (($tsec - $psec) / 3600.0 )]
	dict set ret Route $route

	return $ret
}

##############################################################################
# View

set ::gradi {#f00 #f30 #f60 #f90 #fc0 #ff0 #cf0 #9f0 #6f0 #3f0 #0f0}
set ::maxSpeed 77

proc colorForSpeed {speed} {
	set colorRange [expr $::maxSpeed / [llength $::gradi]]
	set idx [expr int($speed / $colorRange)]
	return [lindex $::gradi $idx]
}

proc drawRoute {name cnvs} {
	set route [dict get $::dataAll $name Route]
	set idx 0
	set pLat [lindex $route 0 1]
	set pLon [lindex $route 0 2]
	foreach point [lrange $route 1 end] {
		lassign $point time lat lon dist dur speed
		$cnvs create line $pLon [expr - $pLat] $lon [expr - $lat] \
			-fill [colorForSpeed $speed] \
			-tags "$name Route Index-$idx"
		set pLat $lat
		set pLon $lon
		incr idx
	}

	# Scale it to be visible
	set bb [$cnvs bbox $name]
	set centerX [expr [lindex $bb 2] - [lindex $bb 0]]
	set centerY [expr [lindex $bb 3] - [lindex $bb 1]]
	$cnvs scale $name $centerX $centerY 2000 2000

	$cnvs config -scrollregion [$cnvs bbox all]
}

proc scrollingCanvas {name} {
	set n [frame $name]

	canvas $n.c \
		-bg gray20 \
		-xscrollincrement 2 \
		-yscrollincrement 2 \
		-xscrollcommand [list $n.x set] \
		-yscrollcommand [list $n.y set]
	scrollbar $n.x -ori hori -command [list $n.c xview]
	scrollbar $n.y -ori vert -command [list $n.c yview]

	grid $n.c $n.y -sticky news
	grid $n.x -sticky ew
	grid rowconfig $n 0 -weight 1
	grid columnconfig $n 0 -weight 1

	$n.c create text 0 0 -fill AliceBlue -text o -font TkFixedFont -tags selection \
		-state hidden

	bind $n.c <Button-2> {puts "[%W canvasx %x] [%W canvasy %y]"}
	bind $n.c <Motion> {setSelection %W [%W canvasx %x] [%W canvasy %y]}
	bind $n.c <ButtonPress-1> {%W scan mark %x %y}
	bind $n.c <B1-Motion> {%W scan dragto %x %y 1}
	bind $n.c <Enter> {focus %W}
	bind $n.c <Key-Prior> {sCZoom %W 1 %x %y}
	bind $n.c <Key-Next> {sCZoom %W -1 %x %y}

	return $n
}

proc setSelection {cnvs x y} {
	$cnvs itemconfigure selection -state hidden

	set item [$cnvs find closest $x $y]
	if {$item eq ""} return
	lassign [$cnvs coords $item] cl ct cr cb
	showRouteStepInfo [$cnvs gettags $item]

	$cnvs itemconfigure selection -state normal
	$cnvs coords selection $cl $ct
}

proc sCZoom {cnvs direction x y} {
	if {$direction > 0} {
		set fact 1.1
	} else {
		set fact [expr 1.0/1.1]
	}
	$cnvs scale all [$cnvs canvasx $x] [$cnvs canvasy $y] $fact $fact
	$cnvs configure -scrollregion [$cnvs bbox all]
}


proc dataSets {name} {
	set n [frame $name]

	set f [frame $n.btns]
	button $f.load -text Load -command loadRoute
	button $f.loag -text Load* -command loadGlob
	button $f.show -text Show -command showRoutes
	button $f.hide -text Hide -command hideRoutes
	button $f.hida -text "Hide All" -command hideAllRoutes

	eval pack [winfo children $f] -fill x -anchor n
	pack $f -side left -anchor n

	set columns {0 Name 0 Start 0 Stop 0 Durration 0 Distance 0 "Avg Speed"}
	pack [scrollingTableList $n.table $columns] -side left -expand yes -fill both

	$n.table columnconfigure 3 -formatcommand {format "%.3f"}
	$n.table columnconfigure 4 -formatcommand {format "%.3f"}
	$n.table columnconfigure 5 -formatcommand {format "%.3f"}

	bind [$n.table bodypath] <Key-s> toggleRoutes

	return $n
}

variable ::dataTable
proc scrollingTableList {name columns} {
	set t [frame $name]

	tablelist::tablelist $t.t \
		-columns $columns \
		-width 100 \
		-showseparators yes \
		-stretch all \
		-listvariable ::dataTable \
		-activestyle frame \
		-stripebackground lightblue \
		-labelcommand tablelist::sortByColumn \
		-yscrollcommand [list $t.y set] \
		-xscrollcommand [list $t.x set] 
	scrollbar $t.y -ori vert -command [list $t.t yview]
	scrollbar $t.x -ori hori -command [list $t.t xview]
	grid $t.t -column 0 -row 0 -sticky news
	grid $t.y -column 1 -row 0 -sticky nes
	grid $t.x -column 0 -row 1 -sticky new
	grid columnconfigure $t 0 -weight 1
	grid rowconfigure $t 0 -weight 1

	rename $t $t.internal
	proc $t {args} [string map [list WIDGET $t] {
		switch [lindex $args 0] {
			add {
				# .table add data
				set data [lrange $args 1 end]
				set idx [lsearch -exact $::dataTable [lindex $data 0]]
				set ::dataTable [lreplace $::dataTable $idx $idx $data]
			}
			default {
				WIDGET.t {*}$args
			}
		}
	}]

	return $t
}

proc infoFrame {name} {
	set n [frame $name]

	set f [frame $n.speedLegend]
	set step [expr $::maxSpeed / [llength $::gradi]]
	set speed 0
	foreach g $::gradi {
		label $f.s$speed -text $speed -fg $g -bg gray20
		incr speed $step
	}
	eval pack [winfo children $f] -fill x -anchor n
	pack $f -side left -anchor n

	set f [frame $n.tags]
	grid [label $f.lroute -text Route] -column 0 -row 0 -sticky nw
	grid [label $f.route -text "" -anchor e] -column 1 -row 0 -sticky ne -columnspan 2

	grid [label $f.llat -text Latitude] -column 0 -row 1 -sticky nw
	grid [label $f.lat -text 0 -width 9 -anchor e] -column 1 -row 1 -sticky ne

	grid [label $f.llon -text Longitude] -column 0 -row 2 -sticky nw
	grid [label $f.lon -text 0 -width 9 -anchor e] -column 1 -row 2 -sticky ne

	grid [label $f.ldur -text Duration] -column 0 -row 3 -sticky nw
	grid [label $f.dur -text 0 -width 9 -anchor e] -column 1 -row 3 -sticky ne
	grid [label $f.udur -text Seconds] -column 2 -row 3 -sticky ne

	grid [label $f.lspeed -text Speed] -column 0 -row 4 -sticky nw
	grid [label $f.speed -text 0 -width 9 -anchor e] -column 1 -row 4 -sticky ne
	grid [label $f.uspeed -text MPH] -column 2 -row 4 -sticky nw

	pack $f -side top -anchor n

	return $n
}

proc updateInfo {inf route lat lon dur speed} {
	$inf.tags.route configure -text $route 
	$inf.tags.lat configure -text [format "%.5f" $lat]
	$inf.tags.lon configure -text [format "%.5f" $lon]
	$inf.tags.dur configure -text $dur 
	$inf.tags.speed configure -text [format "%.3f" $speed]
}

##############################################################################
# Control
variable ::CTable
variable ::CMap
variable ::CInfo

proc loadDataFromFiles {files} {
	set filename [lindex $files 0]
	set name [load_data $filename]

	$::CTable add $name \
		[dict get $::dataAll $name Start] \
		[dict get $::dataAll $name Stop] \
		[dict get $::dataAll $name Durration] \
		[dict get $::dataAll $name Distance] \
		[dict get $::dataAll $name Speed]

	if {[llength $files] > 1} {
		after 1 [list loadDataFromFiles [lrange $files 1 end]]
	}
}

proc loadRoute {} {
	set fps {
		{{xml file} *.xml}
		{{GPX} *.gpx}
		{{All files} *}
	}
	if {[set fns [tk_getOpenFile -parent . -title "Open GPX" -filetypes $fps -multiple yes]] eq ""} {
		return
	}

	loadDataFromFiles $fns
	focus $::CTable
}

proc loadGlob {} {
	set foo "/Documents and Settings/mtilstra/My Documents/Magic Briefcase/Routes to Work"
	set dir [tk_chooseDirectory -title "GPXes" -mustexist yes -parent . -initialdir $foo]
	set glb [glob -directory $dir -nocomplain "*.{xml,gpx}"]
	if {$glb eq ""} return
	loadDataFromFiles $glb
	focus $::CTable
}

proc showRoutes {} {
	foreach idx [$::CTable curselection] {
		drawRoute [lindex $::dataTable $idx 0] $::CMap
	}
}
proc hideRoutes {} {
	foreach idx [$::CTable curselection] {
		$::CMap delete [lindex $::dataTable $idx 0]
	}
}
proc hideAllRoutes {} {
	$::CMap delete Route
}
proc toggleRoutes {} {
	foreach idx [$::CTable curselection] {
		if {[$::CMap gettags [lindex $::dataTable $idx 0]] ne ""} {
			$::CMap delete [lindex $::dataTable $idx 0]
		} else {
			drawRoute [lindex $::dataTable $idx 0] $::CMap
		}
	}
}

proc showRouteStepInfo {tags} {
	set tags [split $tags]
	set idx [lsearch -inline -glob $tags Index-*]
	if {$idx eq ""} return
	set idx [regexp -inline {\d+} $idx]

	set name [lsearch -inline -not -regexp $tags {Route|Index-}]

	set point [lindex [dict get $::dataAll $name Route] $idx]
	lassign $point time lat lon dist dur speed

	updateInfo $::CInfo $name $lat $lon $dur $speed
}

##############################################################################
# Main
proc buildMainWindow {} {
	wm title . "Mapped"

	# Not ideal, but better than hardcoding above.
	set ::CTable .p.b.table
	set ::CMap   .p.p.c.c
	set ::CInfo  .p.p.i

	panedwindow .p -orient vertical -sashrelief groove
	.p add [dataSets .p.b]

	panedwindow .p.p -orient horizontal -sashrelief groove
	.p.p add [scrollingCanvas .p.p.c]
	.p.p add [infoFrame .p.p.i]

	.p add .p.p
	pack .p -fill both -expand yes

	if {![catch {console title}]} {
		console title "Mapped Console"
		bind . <Key-`> {console show}
	}

}

buildMainWindow

